@@ -13,7 +13,7 @@
 #include "pfqhelp.h"
 #include "pfqmessage.h"
 #include "pfregex.h"
-
+#include "pfqconfig.h"
 #include "pfqlib.h"
 
 #define CAT_BUF_SIZE 20*1024
@@ -32,12 +32,17 @@ int MSGMARK;
 char* cat_buf;
 char* regexps;
 
+struct pfql_context_t *pfql_ctx;
+
 struct msg_list_t *msg_list;	/* Array of CURRENTLY SHOWN msg ids */
 int msg_num;			/* Number of messages in the queue */
 
 time_t last_repaint;
 
 int half_delay_time;
+int show_body_win;
+int show_body_always;
+int show_body_fromline;
 
 int wnd_confirm ( const char *action ) {
 	int c, ret;
@@ -45,13 +50,13 @@ int wnd_confirm ( const char *action ) {
 	WINDOW *w = newwin ( 5, 50, (LINES-5)/2, (COLS-50)/2 );
 	wnd_border ( w );
 	
-	if ( pfql_getstatus()->use_colors )
+	if ( pfql_getstatus(pfql_ctx)->use_colors )
 		wbkgd ( w, COLOR_PAIR(CP_DIALOG) );
 
 	nocbreak();
 	cbreak();
 	
-	if ( pfql_getstatus()->wrk_tagged || (pfql_getstatus()->auto_wrk_tagged && pfql_num_tag()) )
+	if ( pfql_getstatus(pfql_ctx)->wrk_tagged || (pfql_getstatus(pfql_ctx)->auto_wrk_tagged && pfql_num_tag(pfql_ctx)) )
 		mvwprintw ( w, 2, 2, "Do you really want to %s tagged messages?", action );
 	else
 		mvwprintw ( w, 2, 2, "Do you really want to %s this message?", action );
@@ -70,7 +75,7 @@ void wnd_getregexp () {
 	char *tmp;
 	
 	WINDOW *w = newwin ( 5, 60, (LINES-5)/2, (COLS-60)/2 );
-	if ( pfql_getstatus()->use_colors )
+	if ( pfql_getstatus(pfql_ctx)->use_colors )
 		wbkgd ( w, COLOR_PAIR(CP_DIALOG) );
 	wnd_border ( w );
 
@@ -87,17 +92,17 @@ void wnd_getregexp () {
 	halfdelay ( half_delay_time*10 );
 	noecho();
 
-	search_mode = SM_ALL;
+	pfql_ctx->search_mode = SM_ALL;
 	if ( !strncmp( regexps, "f:", 2 ) )
-		search_mode = SM_FROM;
+		pfql_ctx->search_mode = SM_FROM;
 	if ( !strncmp( regexps, "t:", 2 ) )
-		search_mode = SM_TO;
+		pfql_ctx->search_mode = SM_TO;
 	if ( !strncmp( regexps, "e:", 2 ) )
-		search_mode = SM_TO | SM_FROM;
+		pfql_ctx->search_mode = SM_TO | SM_FROM;
 	if ( !strncmp( regexps, "s:", 2 ) )
-		search_mode = SM_SUBJ;
+		pfql_ctx->search_mode = SM_SUBJ;
 
-	if ( search_mode != ( SM_ALL ) ) {
+	if ( pfql_ctx->search_mode != ( SM_ALL ) ) {
 		tmp = (char*) malloc ( REGEXP_LEN );
 		strncpy ( tmp, regexps+2, REGEXP_LEN );
 		strncpy ( regexps, tmp, REGEXP_LEN );
@@ -108,7 +113,7 @@ void wnd_getregexp () {
 }
 
 void version() {
-	printf ( "pfqueue - Version %s - (C) 2004, 2005 Stefano Rivoir\n",
+	printf ( "pfqueue - Version %s - (C) 2004, 2005, 2006 Stefano Rivoir\n",
 			VERSION );
 }
 
@@ -131,10 +136,10 @@ void help() {
 			"  Backend API version: %d\n\n"
 			"  Backend ID: %s\n"
 			"  Backend version: %s",
-			pfql_version(), pfql_backend_apiversion(), pfql_backend_id(), pfql_backend_version() );
+			pfql_version(), pfql_backend_apiversion(pfql_ctx), pfql_backend_id(pfql_ctx), pfql_backend_version(pfql_ctx) );
 
 	w = newwin ( 25, 60, (LINES-25)/2,(COLS-60)/2 );
-	if ( pfql_getstatus()->use_colors )
+	if ( pfql_getstatus(pfql_ctx)->use_colors )
 		wbkgd ( w, COLOR_PAIR(CP_DIALOG) );
 
 	nocbreak();
@@ -179,6 +184,8 @@ int goto_msg(int i) {
 	CURMSG  = i;
 	CURROW  = (i-FIRSTMSG)+FIRSTROW;
 
+	show_body_fromline = 0;
+
 	if ( CURROW<FIRSTROW ) {
 		CURROW = FIRSTROW;
 		FIRSTMSG = CURMSG;
@@ -229,34 +236,39 @@ void win_header() {
 
 	wattron ( hwnd, A_BOLD );
 
-	if ( pfql_getstatus()->use_colors )
+	if ( pfql_getstatus(pfql_ctx)->use_colors )
 		wattron ( hwnd, COLOR_PAIR(CP_HEAD) );
 
 	mvwprintw ( hwnd, 0, 1, "Queue: '%s', %d message%s, %d tagged",
-			pfql_queue_name(pfql_getstatus()->cur_queue), msg_num, 
+			pfql_queue_name(pfql_ctx,pfql_getstatus(pfql_ctx)->cur_queue), msg_num, 
 			( msg_num != 1 ? "s" : "" ),
-			pfql_num_tag() );
+			pfql_num_tag(pfql_ctx) );
 	wclrtoeol ( hwnd );
 
 	wattroff ( hwnd, A_REVERSE );
-	if ( pfql_getstatus()->auto_wrk_tagged )
+	if ( pfql_getstatus(pfql_ctx)->auto_wrk_tagged )
 		wattron ( hwnd, A_REVERSE );
-	mvwaddstr ( hwnd, 0, COLS-6, "A" );
+	mvwaddstr ( hwnd, 0, COLS-7, "A" );
 		
 	wattroff ( hwnd, A_REVERSE );
-	if ( pfql_getstatus()->wrk_tagged )
+	if ( pfql_getstatus(pfql_ctx)->wrk_tagged )
 		wattron ( hwnd, A_REVERSE );
-	mvwaddstr ( hwnd, 0, COLS-5, "T" );
+	mvwaddstr ( hwnd, 0, COLS-6, "T" );
 		
 	wattroff ( hwnd, A_REVERSE );
-	if ( pfql_getstatus()->ask_confirm )
+	if ( pfql_getstatus(pfql_ctx)->ask_confirm )
 		wattron ( hwnd, A_REVERSE );
-	mvwaddstr ( hwnd, 0, COLS-4, "C" );
+	mvwaddstr ( hwnd, 0, COLS-5, "C" );
 		
 	wattroff ( hwnd, A_REVERSE );
-	if ( pfql_getstatus()->do_scan )
+	if ( pfql_getstatus(pfql_ctx)->do_scan )
 		wattron ( hwnd, A_REVERSE );
-	mvwaddstr ( hwnd, 0, COLS-3, "S" );
+	mvwaddstr ( hwnd, 0, COLS-4, "S" );
+
+	wattroff ( hwnd, A_REVERSE );
+	if ( show_body_always )
+		wattron ( hwnd, A_REVERSE );
+	mvwaddstr ( hwnd, 0, COLS-3, "B" );
 
 	wattroff ( hwnd, A_REVERSE );
 	wattron ( hwnd, A_BOLD );
@@ -270,13 +282,21 @@ void win_header() {
 
 // Sizes the window according to the (new) screen size
 void win_resize () {
-
-	if ( mwndb )
-		delwin ( mwndb );
-	if ( lwnd )
-		delwin ( lwnd );
+	int mwnd_tot;
+	int bwnd_tot;
+	int lwnd_tot;
+	
+	bwnd_tot = bwnd_height + 2;
 	if ( mwnd )
 		delwin ( mwnd );
+	if ( mwndb )
+		delwin ( mwndb );
+	if ( bwnd )
+		delwin ( bwnd );
+	if ( bwndb )
+		delwin ( bwndb );
+	if ( lwnd )
+		delwin ( lwnd );
 	if ( hwnd )
 		delwin ( hwnd );
 	if ( bline )
@@ -286,36 +306,53 @@ void win_resize () {
 
 	last_repaint = 0;
 	mainwnd = initscr();
-	MAXROW = LINES-8-1-2;
 	FIRSTROW = 0;
 
 	if ( !has_colors() || start_color()==ERR )
-		pfql_getstatus()->use_colors = 0;
+		pfql_getstatus(pfql_ctx)->use_colors = 0;
 
 	bline = (char*)malloc(sizeof(char)*COLS+1);
 	memset ( bline, ' ', sizeof(char)*COLS+1 );
 	*(bline+COLS) = 0;
 	msg_list = (struct msg_list_t*)malloc(sizeof(struct msg_list_t)*(LINES-8-2));
 
-	hwnd  = newwin ( 2, COLS, 0, 0 );
-	lwnd  = newwin ( LINES-8-2, COLS, 2, 0 );
-	mwndb = newwin ( 8, COLS, LINES-8, 0 );
-	mwnd  = newwin ( 8-2, COLS-2, LINES-8+1, 1 );
+	mwnd_tot = 7;
+	lwnd_tot = LINES-mwnd_tot-2;
+	if ( show_body_win )
+		lwnd_tot -= bwnd_height;
 
-	if ( pfql_getstatus()->use_colors ) {
+	MAXROW = LINES-(mwnd_tot)-1-2;
+	if ( show_body_win )
+		MAXROW -= bwnd_height;
+		
+	hwnd  = newwin ( 2, COLS, 0, 0 );
+	lwnd  = newwin ( lwnd_tot, COLS, 2, 0 );
+	mwndb = newwin ( mwnd_tot, COLS, lwnd_tot+2, 0 );
+	mwnd  = newwin ( mwnd_tot-2, COLS-2, lwnd_tot+3, 1 );
+	if ( show_body_win ) {
+		bwndb = newwin ( bwnd_height,   COLS,   lwnd_tot+2+mwnd_tot,   0 );
+		bwnd  = newwin ( bwnd_height-2, COLS-2, lwnd_tot+2+mwnd_tot+1, 1 );
+	}
+
+	if ( pfql_getstatus(pfql_ctx)->use_colors ) {
 		init_pair ( CP_MSG, COLOR_WHITE, COLOR_MAGENTA );
 		init_pair ( CP_HEAD,COLOR_WHITE, COLOR_BLUE );
 		init_pair ( CP_TAG, COLOR_YELLOW, COLOR_BLACK );
 		init_pair ( CP_DIALOG, COLOR_WHITE, COLOR_RED );
 		init_pair ( CP_OPTDIS, COLOR_RED, COLOR_BLUE );
 		wbkgd ( mwndb, COLOR_PAIR(CP_HEAD) );
+		if ( show_body_win )
+			wbkgd ( bwndb, COLOR_PAIR(CP_HEAD) );
 		wbkgd ( hwnd, COLOR_PAIR(CP_HEAD) );
 	}
 
-	if ( pfql_getstatus()->use_envelope )
+	if ( pfql_getstatus(pfql_ctx)->use_envelope )
 		wnd_clear ( mwndb, "Message (from envelope)" );
 	else
 		wnd_clear ( mwndb, "Message (from headers)" );
+	
+	if ( show_body_win )
+		wnd_clear ( bwndb, "Message body" );
 
 	COL1 = 20;
 	COL2 = COL1+( COLS - COL1 ) / 2-2;
@@ -324,6 +361,8 @@ void win_resize () {
 	halfdelay(half_delay_time*10);
 
 	wnoutrefresh ( mwndb );
+	if ( show_body_win )
+		wnoutrefresh ( bwndb );
 
 	page_step = MAXROW ;
 
@@ -331,8 +370,7 @@ void win_resize () {
 }
 
 // Displays the bottom window
-void msg_show () {
-
+void msg_show_headers () {
 	struct msg_t *msg;
 	
 	if ( !msg_num ) {
@@ -341,12 +379,12 @@ void msg_show () {
 		return;
 	}
 	
-	msg = pfql_msg( msg_list[CURROW].id );
+	msg = pfql_msg( pfql_ctx, msg_list[CURROW].id );
 	if ( !msg )
 		return;
 	
-	pfql_retr_status ( msg->id );
-	pfql_retr_headers( msg->id );
+	pfql_retr_status ( pfql_ctx, msg->id );
+	pfql_retr_headers( pfql_ctx, msg->id );
 
 	mvwaddstr ( mwnd, 0, 0, "Message:" );
 	wclrtobot ( mwnd );
@@ -364,6 +402,9 @@ void msg_show () {
 	mvwaddstr ( mwnd, 4, 0, "Status:" );
 	mvwaddstr ( mwnd, 5, 0, " " );
 	mvwaddstr ( mwnd, 4, 8, msg->stat );
+
+	if ( show_body_always && show_body_win )
+		msg_show_body(0);
 }
 
 // Returns the message displayed at row r
@@ -381,13 +422,13 @@ void show_cursor() {
 	// Current row
 	i = msg_at(CURROW);
 	mvwinnstr ( lwnd, CURROW, 0, buf, COLS );
-	if ( pfql_getstatus()->use_colors ) {
+	if ( pfql_getstatus(pfql_ctx)->use_colors ) {
 		wattron ( lwnd, COLOR_PAIR(CP_MSG) );
 	} else {
 		wattron ( lwnd, A_REVERSE );
 	}
 
-	msg = pfql_msg_at(i);
+	msg = pfql_msg_at(pfql_ctx,i);
 	if ( msg && msg->tagged )
 		wattron ( lwnd, A_BOLD );
 	
@@ -400,11 +441,11 @@ void show_cursor() {
 	if ( PREVROW!=CURROW ) {
 		i = msg_at(PREVROW);
 		mvwinnstr ( lwnd, PREVROW, 0, buf, COLS );
-		msg = pfql_msg_at(i);
+		msg = pfql_msg_at(pfql_ctx,i);
 		if ( msg ) {
 			if ( msg->tagged )
 				wattron ( lwnd, A_BOLD );
-			if ( msg->tagged && pfql_getstatus()->use_colors )
+			if ( msg->tagged && pfql_getstatus(pfql_ctx)->use_colors )
 				wattron ( lwnd, COLOR_PAIR(CP_TAG) );
 		}
 		mvwaddstr ( lwnd, PREVROW, 0, buf );
@@ -415,22 +456,17 @@ void show_cursor() {
 	wattroff ( lwnd, A_BOLD );
 	wattroff ( lwnd, A_REVERSE );
 
-	/*
-	mvwprintw ( hwnd, 1, 30, "lc:%d lu:%d c:%d", pfql_queue_last_changed(), last_repaint,
-			pfql_msg_at(CURROW)->hcached );
-	*/
 	wrefresh ( hwnd );
 
 	wmove ( lwnd, CURROW, 0 );
 
 	PREVROW = CURROW;
 	
-	msg_show();
+	msg_show_headers();
 	wrefresh ( mwnd );
 }
 
-void msg_cat() {
-	WINDOW *w;
+void msg_cat_body( WINDOW *pw, WINDOW* bw, int wl, int wc, int ctrls ) {
 	char lines[MSG_MAX_ROWS][MSG_MAX_COLS];
 	int j, i;
 	int c;
@@ -441,11 +477,10 @@ void msg_cat() {
 	int buflen;
 	int maxcol;
 
-	maxcol = COLS-4-2;
-	
+	maxcol = wc-2;
 	memset ( lines, 0, sizeof(lines) );
 	
-	buflen = pfql_retr_body ( msg_list[CURROW].id, cat_buf, CAT_BUF_SIZE );
+	buflen = pfql_retr_body ( pfql_ctx, msg_list[CURROW].id, cat_buf, CAT_BUF_SIZE );
 	if ( buflen == PFQL_MSGNOTEX )
 		return;
 
@@ -466,31 +501,39 @@ void msg_cat() {
 		msglines++;
 	}
 
-	w = newwin ( LINES-4, COLS-4, 2, 2 );
-	keypad ( w, TRUE );
-	wnd_border (w);
-	pgstep = LINES-4-2-2;
+	keypad ( pw, TRUE );
+	pgstep = wl;
 	
-	wattron ( w, A_BOLD );
-	mvwaddstr ( w, 0, 2, "Details for" );
-	mvwaddstr ( w, 0, 14, pfql_msg_at(CURMSG)->path );
-	wattroff ( w, A_BOLD );
+	if ( bw ) {
+		wattron ( bw, A_BOLD );
+		mvwaddstr ( bw, 0, 2, "Body of" );
+		mvwaddstr ( bw, 0, 10, pfql_msg_at(pfql_ctx,CURMSG)->path );
+		wattroff ( bw, A_BOLD );
+		wrefresh ( bw );
+	}
 
 	c = 0;
-	fline = 0;
+	if ( ctrls )
+		fline = 0;
+	else
+		fline = show_body_fromline;
 
 	while ( c!='q' && c!='Q' ) {
 		j = fline;
-		i = 2;
-		while ( j<MSG_MAX_ROWS && i < LINES-4-1 ) {
-			mvwaddnstr ( w, i, 2, lines[j++], COLS-4-5 );
-			wclrtoeol ( w );
-			mvwaddch ( w, i++, COLS-4-1, ACS_VLINE );
+		i = 0;
+		while ( j<MSG_MAX_ROWS && i < wl ) {
+			mvwaddnstr ( pw, i++, 2, lines[j++], maxcol );
+			wclrtoeol ( pw );
+			//mvwaddch ( pw, i++, COLS-4-1, ACS_VLINE );
 		}
-		wrefresh ( w );
+		wrefresh ( pw );
 
 no_fill:
-		c = wgetch (w);
+		if ( ctrls )
+			c = wgetch (pw);
+		else
+			c = 'q';
+
 		switch (c) {
 		case 'g':
 		case KEY_HOME:
@@ -527,14 +570,28 @@ no_fill:
 			break;
 		}
 	}
-	delwin ( w );
 
 	halfdelay ( half_delay_time*10 );
 }
 
+void msg_show_body(int force_own) {
+	WINDOW *w1, *w2;
+	if ( show_body_win && !force_own )
+		msg_cat_body ( bwnd, bwndb, bwnd_height-2, COLS-2, 0 );
+	else {
+		w1 = newwin ( LINES-4, COLS-4, 2, 2 );
+		w2 = newwin ( LINES-6, COLS-6, 3, 3 );
+		wnd_border ( w1 );
+		wrefresh ( w1 );
+		msg_cat_body ( w2, w1, LINES-6, COLS-6, 1 );
+		delwin ( w2 );
+		delwin ( w1 );
+	}
+}
+
 // Returns 1 if queue needs repaint
 int queue_needs_repaint () {
-	return ( pfql_queue_last_changed() >= last_repaint );
+	return ( pfql_queue_last_changed(pfql_ctx) >= last_repaint );
 }
 
 // Set repaint flag
@@ -542,7 +599,7 @@ void queue_force_repaint(int f) {
 	if ( f )
 		last_repaint = 0;
 	else
-		last_repaint = pfql_queue_last_changed();
+		last_repaint = pfql_queue_last_changed(pfql_ctx);
 }
 
 // Displays the list of messages
@@ -554,7 +611,7 @@ struct msg_t *msg;
 	j = FIRSTROW;
 	i = FIRSTMSG;
 
-	msg_num = pfql_num_msg();
+	msg_num = pfql_num_msg(pfql_ctx);
 	
 	//if ( dig_limit )
 	//	show_start = time(NULL);
@@ -566,16 +623,16 @@ struct msg_t *msg;
 
 		if ( i<msg_num ) {
 
-			msg = pfql_msg_at(i);
+			msg = pfql_msg_at(pfql_ctx,i);
 			if ( !msg )
 				return;
 			strcpy ( msg_list[j].id, msg->id );
 
-			pfql_retr_headers ( msg->id );
+			pfql_retr_headers ( pfql_ctx,msg->id );
 
 			if ( msg->tagged ) {
 				wattron ( lwnd, A_BOLD );
-				if ( pfql_getstatus()->use_colors )
+				if ( pfql_getstatus(pfql_ctx)->use_colors )
 					wattron ( lwnd, COLOR_PAIR(CP_TAG) );
 			}
 			mvwaddnstr  ( lwnd, j, 0, bline, COLS );
@@ -608,6 +665,8 @@ struct msg_t *msg;
 // Init frontend
 int fe_init() {
 	cat_buf = 0;
+	bwnd_height = 10;
+
 	cat_buf = (char*)malloc(CAT_BUF_SIZE);
 	if ( !cat_buf )
 		return 0;
@@ -617,7 +676,7 @@ int fe_init() {
 	if ( !regexps )
 		return 0;
 
-	mainwnd = mwndb = lwnd = mwnd = hwnd = NULL;
+	mainwnd = mwndb = lwnd = mwnd = hwnd = bwnd = bwndb = NULL;
 	bline = 0;
 	msg_list = 0;
 	return 1;
@@ -635,7 +694,7 @@ void fe_close() {
 }
 
 void queue_change(int q) {
-	pfql_set_queue(q);
+	pfql_set_queue(pfql_ctx,q);
 	CURMSG = 0;
 	goto_first();
 	win_header();
@@ -644,9 +703,9 @@ void queue_change(int q) {
 void msg_action(const char* id, int act) {
 	if ( !msg_num )
 		return;
-	if ( (pfql_getstatus()->ask_confirm && wnd_confirm(a_names[act])) ||
-	     (!pfql_getstatus()->ask_confirm) )
-		pfql_msg_action(id, act);
+	if ( (pfql_getstatus(pfql_ctx)->ask_confirm && wnd_confirm(a_names[act])) ||
+	     (!pfql_getstatus(pfql_ctx)->ask_confirm) )
+		pfql_msg_action(pfql_ctx, id, act);
 	queue_force_repaint(1);
 }
 
@@ -694,7 +753,7 @@ int i;
 			win_resize();
 			break;
 		case 'e':
-			pfql_toggle_envelope();
+			pfql_toggle_envelope(pfql_ctx);
 			queue_force_repaint ( 1 );
 			break;
 		case 'd':
@@ -723,27 +782,37 @@ int i;
 			queue_change ( Q_ACTIVE );
 			break;
 		case 'a':
-			pfql_tag_all();
+			pfql_tag_all(pfql_ctx);
 			queue_force_repaint ( 1 );
 			win_header();
 			break;
-		case '.':
+		case 'm':
 			MSGMARK = CURMSG;
-			pfql_msg_tag( msg_list[CURROW].id);
+			pfql_msg_tag( pfql_ctx, msg_list[CURROW].id);
+			break;
+		case 'b':
+			show_body_win = !show_body_win;
+			win_resize();
+			if ( show_body_win )
+				msg_show_body(0);
+			break;
+		case 'B':
+			show_body_always = !show_body_always;
+			win_header();
 			break;
 		case 32:
 		case 't':
 			if ( !msg_num )
 				break;
 			if ( MSGMARK==-1 ) {
-				pfql_msg_toggletag( msg_list[CURROW].id );
+				pfql_msg_toggletag( pfql_ctx,msg_list[CURROW].id );
 			} else {
 				if ( CURMSG>=MSGMARK ) {
 					for ( i=MSGMARK; i<=CURMSG; i++ )
-						pfql_msg_tag(pfql_msg_at(i)->id);
+						pfql_msg_tag(pfql_ctx,pfql_msg_at(pfql_ctx,i)->id);
 				} else {
 					for ( i=CURMSG; i<=MSGMARK; i++ ) 
-						pfql_msg_tag(pfql_msg_at(i)->id);
+						pfql_msg_tag(pfql_ctx,pfql_msg_at(pfql_ctx,i)->id);
 				}
 			}
 			MSGMARK = -1;
@@ -754,63 +823,91 @@ int i;
 			break;
 		case 'T':
 			wnd_getregexp();
-			pfql_msg_searchandtag(regexps);
+			pfql_msg_searchandtag(pfql_ctx,regexps);
 			queue_force_repaint ( 1 );
 			win_header();
 			break;
 		case 'u':
-			pfql_tag_none();
+			pfql_tag_none(pfql_ctx);
 			win_header();
 			queue_force_repaint ( 1 );
 			break;
 		case ':':
-			pfql_getstatus()->auto_wrk_tagged = !pfql_getstatus()->auto_wrk_tagged;
+			pfql_getstatus(pfql_ctx)->auto_wrk_tagged = !pfql_getstatus(pfql_ctx)->auto_wrk_tagged;
 			win_header();
 			break;
 		case ';':
-			pfql_getstatus()->wrk_tagged = !pfql_getstatus()->wrk_tagged;
+			pfql_getstatus(pfql_ctx)->wrk_tagged = !pfql_getstatus(pfql_ctx)->wrk_tagged;
 			win_header();
 			break;
 		case '/':
 			wnd_getregexp();
-			i = pfql_msg_search(regexps);
+			i = pfql_msg_search(pfql_ctx,regexps);
 			goto_msg ( i );
 			queue_force_repaint ( 1 );
 			break;
 		case 'n':
-			i = pfql_msg_searchnext(regexps);
+			i = pfql_msg_searchnext(pfql_ctx,regexps);
 			goto_msg ( i );
 			break;
 		case 'p':
-			pfql_msg_searchprev(regexps);
+			pfql_msg_searchprev(pfql_ctx,regexps);
 			break;
 		case '?':
 			help();
 			queue_force_repaint ( 1 );
 			break;
 		case 'c':
-			pfql_getstatus()->ask_confirm = !pfql_getstatus()->ask_confirm;
+			pfql_getstatus(pfql_ctx)->ask_confirm = !pfql_getstatus(pfql_ctx)->ask_confirm;
 			win_header();
 			break;
 		case '-':
-			pfql_getstatus()->do_scan = !pfql_getstatus()->do_scan;
+			pfql_getstatus(pfql_ctx)->do_scan = !pfql_getstatus(pfql_ctx)->do_scan;
 			win_header();
 			break;
 		case '+':
-			pfql_getstatus()->use_colors = !pfql_getstatus()->use_colors;
+			pfql_getstatus(pfql_ctx)->use_colors = !pfql_getstatus(pfql_ctx)->use_colors;
 			queue_force_repaint ( 1 );
 			win_resize();
 			win_header();
 			break;
+		case '>':
+			if ( bwnd_height<LINES-10 )
+				bwnd_height++;
+			win_resize();
+			win_header();
+			break;
+		case '<':
+			if ( bwnd_height>5 )
+				bwnd_height--;
+			win_resize();
+			win_header();
+			break;
+		case ',':
+			if ( show_body_fromline && show_body_win )
+				show_body_fromline--;
+			msg_show_body(0);
+			break;
+		case '.':
+			show_body_fromline++;
+			msg_show_body(0);
+			break;
 		case 13:
 		case 10:
+			msg_show_body(0);
+			if ( !show_body_win ) {
+				win_resize();
+				queue_force_repaint ( 1 );
+				msg_show_headers();
+			}
+			break;
 		case 's':
 		case 'S':
 			if ( msg_num ) {
-				msg_cat();
+				msg_show_body(1);
 				win_resize();
 				queue_force_repaint ( 1 );
-				msg_show();
+				msg_show_headers();
 			}
 			break;
 		}
@@ -827,13 +924,19 @@ int main ( int argc, char** argv ) {
 		exit (-3);
 	}
 
-	if ( pfql_init()!=PFQL_OK ) {
+	pfql_ctx = 0;
+	if ( pfql_context_create (&pfql_ctx)!=PFQL_OK ) {
+		fprintf ( stderr, "pfquque: cannot pfql_create_context!\n" );
+		exit (-1);
+	}
+
+	if ( pfql_init(pfql_ctx)!=PFQL_OK ) {
 		fprintf ( stderr, "pfqueue: cannot pfql_init!\n" );
 		exit (-1);
 	}
 	
 	if ( !fe_init() ) {
-		pfql_close();
+		pfql_context_destroy(pfql_ctx);
 		fe_close();
 		fprintf ( stderr, "pfqueue: cannot fe_init!\n" );
 		exit (-2);
@@ -847,27 +950,31 @@ int main ( int argc, char** argv ) {
 	 */
 
 	// Some defaults
-	regexp = 0;
+	pfql_ctx->regexp = 0;
 	half_delay_time = 1;
+	show_body_win = 0;
+	show_body_always = 1;
+
+	pfq_read_config ( pfql_ctx );
 
 	opt = 0;
 	while ( opt!=-1 ) {
 		opt = getopt ( argc, argv, "p:c:r:B:b:q:s:m:l:ehvn" );
 		switch (opt) {
 			case 'p':
-				snprintf ( pfql_getconf()->backend_progs, pfql_getconf()->max_char, "%s", optarg );
+				snprintf ( pfql_getconf(pfql_ctx)->backend_progs, pfql_getconf(pfql_ctx)->max_char, "%s", optarg );
 				break;
 			case 'c':
-				snprintf ( pfql_getconf()->backend_config, pfql_getconf()->max_char, "%s", optarg );
+				snprintf ( pfql_getconf(pfql_ctx)->backend_config, pfql_getconf(pfql_ctx)->max_char, "%s", optarg );
 				break;
 			case 'r':
-				snprintf ( pfql_getconf()->remote_host, pfql_getconf()->max_char, "%s", optarg );
+				snprintf ( pfql_getconf(pfql_ctx)->remote_host, pfql_getconf(pfql_ctx)->max_char, "%s", optarg );
 				break;
 			case 'B':
-				snprintf ( pfql_getconf()->backends_path, pfql_getconf()->max_char, "%s", optarg );
+				snprintf ( pfql_getconf(pfql_ctx)->backends_path, pfql_getconf(pfql_ctx)->max_char, "%s", optarg );
 				break;
 			case 'b':
-				snprintf ( pfql_getconf()->backend_name, pfql_getconf()->max_char, "%s", optarg );
+				snprintf ( pfql_getconf(pfql_ctx)->backend_name, pfql_getconf(pfql_ctx)->max_char, "%s", optarg );
 				break;
 			case 'q':
 				mm = atoi ( optarg );
@@ -887,7 +994,7 @@ int main ( int argc, char** argv ) {
 				default:
 					mm = 0;
 				}
-				pfql_getconf()->initial_queue = mm;
+				pfql_getconf(pfql_ctx)->initial_queue = mm;
 				break;
 			case 'h':
 				usage();
@@ -901,7 +1008,7 @@ int main ( int argc, char** argv ) {
 				mm = atoi (optarg);
 				if ( mm < 5 )
 					mm = 5;
-				pfql_getconf()->msg_max = mm;
+				pfql_getconf(pfql_ctx)->msg_max = mm;
 				break;
 			case 's':
 				half_delay_time = atoi ( optarg );
@@ -916,22 +1023,25 @@ int main ( int argc, char** argv ) {
 					mm = 300;
 				if ( mm < 1 )
 					mm = 1;
-				pfql_getconf()->scan_limit = mm;
+				pfql_getconf(pfql_ctx)->scan_limit = mm;
 				break;
 			case 'e':
-				pfql_getstatus()->use_envelope = TRUE;
+				pfql_getstatus(pfql_ctx)->use_envelope = TRUE;
 				break;
 			case 'n':
-				pfql_getstatus()->use_colors = 0;
+				pfql_getstatus(pfql_ctx)->use_colors = 0;
+				break;
+			case 'd':
+				pfql_getconf(pfql_ctx)->scan_delay = atoi(optarg);
 				break;
 			case '?':
 				goto do_exit;
 		}
 	}
 
-	if ( pfql_start() != PFQL_OK ) {
+	if ( pfql_start(pfql_ctx) != PFQL_OK ) {
 		fprintf ( stderr, "pfqueue: cannot start pfqlib. See syslog for details\n" );
-		pfql_close();
+		pfql_context_destroy(pfql_ctx);
 		exit (-4);
 	}
 	
@@ -940,7 +1050,8 @@ int main ( int argc, char** argv ) {
 	main_loop();
 
 do_exit:
-	pfql_close();
+	if ( pfql_ctx )
+		pfql_context_destroy ( pfql_ctx );
 	fe_close();
 	endwin();
 
