@@ -9,7 +9,7 @@
 	"  h) Hold message\n"\
 	"  r) Requeue message\n"\
 	"  l) Release message in 'hold' queue\n"\
-	"  .) Mark message as tag block start\n"\
+	"  m) Mark message as tag block start\n"\
 	"  t) Tag/Untag message (up to mark if present)\n"\
 	"  a) Tag all messages\n"\
 	"  u) Untag all messages\n"\
@@ -31,5 +31,8 @@
 	"  1) Deferred queue\n"\
 	"  2) Active queue\n"\
 	"  3) Incoming queue\n"\
-	"  4) Hold queue\n"
-;
+	"  4) Hold queue\n"\
+	"  >) Increase body window height\n"\
+	"  <) Decrease body window height\n"\
+	"  ,) Scroll body window up\n"\
+	"  .) Scroll body window down\n"
